#
# header.rb
#
#     Copyright(c) 1998-1999 Minero Aoki
#     aamine@dp.u-netsurf.ne.jp
#

require 'extmod'
require 'mailp'
require 'bencode'
require 'addr'
require 'timezone'


class HeaderField

  def initialize( fname, str )
    @name = (fname.must String)
    @body = (str.must String)
    @body.strip!
    parse
  end


  class << self

    alias hf_original_new new

    def new( fname, str )
      if self == HeaderField then

        temp = fname.downcase
        if /\Ax-/ === temp then
          cls = StringH
        else
          cls = (STR2CLASS[temp] || UnknownH)
        end

        cls.new( fname, str )
      else
        hf_original_new( fname, str )
      end
    end
  
  end


  def hash() @name.hash end

  def eql?( oth )
    oth.type == self.type and @name == oth.name
  end

  def ==( oth ) eql? oth end
    

  abstract :body

  def name() @name.dup end

  def to_s( sep )
    @name + ': ' + @body
  end

  def inspect
    @name + ': ' + self.body
  end

end



# HString  ---------------------------------------------


class StringH < HeaderField

  def initialize( fname, body )
    super( fname, body )
    @deced = nil
  end

  def to_s( sep )
    if @deced then
      return Bencode.encode( @name + ': ' + @deced, 74, sep )
    else
      return @name + ': ' + @body
    end
  end

  def body
    unless @deced then
      @deced = Bencode.decode( @body )
    end
    return @deced
  end

  def body=( str )
    str.must String
    @deced = Bencode.decode( str )

    return @deced
  end

  def eql?( oth )
    super and
    @body == oth.body
  end


  private

  def parse
  end

end


# struct type -----------------------------------------

class StructH < HeaderField

  attr :comments

  abstract :body

  def eql?( oth )
    super and
    @body == oth.body
  end


  private


  def init
    @comments = []
  end

  def parse
    unless @parsing then
      @parsing = true
      init
      Mailp.parse( self, @body )
      @parsing = false
    end
  end

end


# unknown type ----------------------------------------

class UnknownH < StructH

  def initialize( fname, body )
    super( fname, body )
    @deced = nil
  end

  def body
    unless @deced then
      @deced = Bencode.decode( @body )
    end

    return @deced
  end

  def body=( str )
    @deced = Bencode.decode( str.must String )
  end

  def eql?( oth )
    super and
    @deced == oth.body
  end


  private

  def init
    super
  end

  def parse
    init
  end

end


# date type -------------------------------------------

class DateH < StructH

  property :date, Time, true, false

  def date=( arg )
    arg.must Time
    @date = arg.localtime
  end


  def DateH.t2s( tim )
    tim.must Time

    day = tim.mday.to_s
    ret = tim.strftime( "%a, #{day} %b %Y %X " )
    ret << Time.zone_s2s( tim.zone )

    return ret
  end

  def body
    DateH.t2s( @date )
  end

  def eql?( oth )
    super and
    self.date == oth.date
  end


  private

  def init
    super
  end

end


# return path type -------------------------------------

class RetpathH < StructH

  property :routes, Array,  true, false
  property :addr,   String, true, true

  def body
    str = '<'
    if @routes.size > 0 then
      str << @routes.collect{|i| '@' + i }.join(',') << ':'
    end
    str << @addr << '>'
  end

  def eql?( oth )
    super and
    self.addr == oth.addr and
    self.routes == oth.routes
  end


  private

  def init
    super
    @routes = []
  end

end


# saddr type -------------------------------------------

class SaddrH < StructH
  
  property :addr, :Mbox, true, true

  def body
    @addr.to_s
  end

  def eql?( oth )
    super and
    self.addr == oth.addr
  end


  private

  def init
    super
  end

end


# maddr type -----------------------------------------

class MaddrH < StructH

  property :addrs, Array, true, false

  def body
    @addrs.collect{|a| a.inspect}.join(', ')
  end

  def eql?( oth )
    super and
    self.addrs == oth.addrs
  end


  private

  def init
    super
    @addrs = []
  end

end


# ref type -------------------------------------------

class RefH < StructH

  property :refs, Array, true, false

  def body
    @refs.join(' ')
  end

  def eql?( oth )
    super and
    self.refs == oth.refs
  end

  def each_msgid
    @refs.each do |i|
      yield i if MsgidH.msgid? i
    end
  end

  def each_phrase
    @refs.each do |i|
      yield i unless MsgidH.msgid? i
    end
  end


  private

  def init
    super
    @refs = []
  end

end


# key type -------------------------------------------

class KeyH < StructH

  property :keys, Array, true, false

  def body
    @keys.join(', ')
  end

  def eql?( oth )
    super and
    self.keys == oth.keys
  end


  private

  def init
    super
    @keys = []
  end

end


# received type ---------------------------------------


class RecvH < StructH

  property :from,  [ :String, :NilClass ], true, true
  property :by,    [ :String, :NilClass ], true, true
  property :via,   [ :String, :NilClass ], true, true
  property :with,  [ :Array,  :NilClass ], true, false
  property :msgid, [ :String, :NilClass ], true, true
  property :ford,  [ :Mbox,   :NilClass ], true, true
  property :date,  [ :Time,   :NilClass ], true, true

  def body
    ret = ''

    ret << 'from ' << @from << ' ' if @from
    ret << 'by '   << @by   << ' ' if @by
    ret << 'via '  << @via  << ' ' if @via

    @with.each do |i| ret << 'with ' << i << ' ' end

    ret << 'id '   << @msgid << ' ' if @msgid
    ret << 'for '  << @ford.to_s    if @ford

    if @date then
      ret.strip!
      ret << '; '
      ret << DateH.t2s( @date )
    end

    return ret
  end

  def eql?( oth )
    super and
    self.from  == oth.from and
    self.by    == oth.by and
    self.via   == oth.via and
    self.with  == oth.with and
    self.msgid == oth.msgid and
    self.ford  == oth.ford and
    self.date  == oth.date
  end


  private

  def init
    super
    @with = []
  end

end


# message-id type  ----------------------------------------------------

class MsgidH < StructH

  property :msgid, :String, true, false

  def msgid=( mid )
    @msgid = (mid.must String)
    if @parsing then
      @msgid = mid
    else
      MSGID === mid or
        raise ArgumentError, "wrong message-ID format: #{mid}"
    end
  end

  MSGID = /<[^\@>]+\@[^>\@]+>/o

  def MsgidH.msgid?( str )
    if MSGID === (str.must String) then
      return $&
    else
      return nil
    end
  end

  def body
    @msgid.dup
  end

  def eql?( oth )
    super and
    self.msgid == oth.msgid
  end


  private

  def init
    super
  end

end



# encrypted type ------------------------------------------------------

class EncH < StructH

  property :encrypter,   :String,              true, true
  property :keyword,   [ :String, :NilClass ], true, true

  def body
    if @key then
      @encrypter + ', ' + @key
    else
      @encrypter.dup
    end
  end

  def eql?( oth )
    super and
    self.encrypter == oth.encrypter and
    self.keyword   == oth.keyword
  end


  private

  def init
    super
  end

end



# version type -----------------------------------------

class VersionH < StructH

  property :major, :Integer, true, true
  property :minor, :Integer, true, true

  def body
    "#{@major}.#{@minor}"
  end

  def eql?( oth )
    super and
    self.major == oth.major and
    self.minor == oth.minor
  end


  private

  def init
    super
  end

end



# content type  ---------------------------------------


class CTypeH < StructH

  property :main,   :String, true, true
  property :sub,    :String, true, true
  property :params, :Hash,   true, false

  def []( key )
    params[key]
  end

  def body
    ret = "#{@main}/#{@sub}"
    params.each_pair do |k,v|
      ret << '; ' << k << ' = ' << v
    end

    return ret
  end
  
  def eql?( oth )
    super and
    self.main == oth.main and
    self.sub  == oth.sub  and
    self.params == oth.params
  end


  private

  def init
    super
    @params = {}
  end

end


# encoding type  ----------------------------


class CEncodingH < StructH

  property :encoding, :String, true, false

  def encoding=( enc )
    @encoding = (enc.must String)
  end

  def body
    @encoding.dup
  end
  
  def eql?( oth )
    super and
    self.encoding == oth.encoding
  end


  private

  def init
    super
  end

end



# disposition type ----------------------------------

class CDispositionH < StructH

  property :disposition, :String, true, true
  property :params,      :Hash,   true, false

  def []( key )
    params[key]
  end

  def body
    ret = @disposition.dup
    @params.each_pair do |k,v|
      ret << '; ' << k << ' = ' << v
    end
    
    return ret
  end

  def eql?( oth )
    super and
    self.disposition == oth.disposition and
    self.params == oth.params
  end
  

  private

  def init
    super
    @params = {}
  end
    
end


# ---------------------------------------------------

class HeaderField   # backward definition

  STR2CLASS = {
    'date'                      => DateH,
    'resent-date'               => DateH,
    'received'                  => RecvH,
    'return-path'               => RetpathH,
    'sender'                    => SaddrH,
    'resent-sender'             => SaddrH,
    'to'                        => MaddrH,
    'cc'                        => MaddrH,
    'bcc'                       => MaddrH,
    'from'                      => MaddrH,
    'reply-to'                  => MaddrH,
    'resent-to'                 => MaddrH,
    'resent-cc'                 => MaddrH,
    'resent-bcc'                => MaddrH,
    'resent-from'               => MaddrH,
    'resent-reply-to'           => MaddrH,
    'message-id'                => MsgidH,
    'resent-message-id'         => MsgidH,
    'content-id'                => MsgidH,
    'in-reply-to'               => RefH,
    'references'                => RefH,
    'keywords'                  => KeyH,
    'encrypted'                 => EncH,
    'mime-version'              => VersionH,
    'content-type'              => CTypeH,
    'content-transfer-encoding' => CEncodingH,
    'content-disposition'       => CDispositionH,
    'subject'                   => StringH,
    'comments'                  => StringH,
    'content-description'       => StringH
  }

end
